#include "coding.hh"
#include "../filesystem/filesystem_common.hh"
#include "../filesystem/filesystem_utils.hh"
#include "../cache/cache.hh"
#include "../gui/diskusage_report.hh"

#include <stdio.h>
#include <stdlib.h>
#include <fuse.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>

#define AbnormalError() printf("%s %s %d\n", __FILE__, __func__,__LINE__);

//NCFS context state
extern struct ncfs_state* NCFS_DATA;
extern FileSystemLayer* fileSystemLayer;
extern CacheLayer* cacheLayer;
extern DiskusageReport* diskusageLayer;


/*************************************************************************
 * Private functions
 *************************************************************************/

/*************************************************************************
 * Galios Field functions
 *************************************************************************/

/*
 * gf_gen_tables: Create logarithm and inverse logairthm tables for computing Q parity
 *
 * @param s: field size (2 power s)
 *
 * @return: 0
 */

int CodingLayer::gf_gen_tables(int s)
{
        unsigned int b, index, gf_elements;

        //convert s to the number of elements for the table
        gf_elements = 1 << s;
        this->gflog = (unsigned short*) malloc(sizeof(unsigned short)*gf_elements);
        this->gfilog = (unsigned short*) malloc(sizeof(unsigned short)*gf_elements);
        b = 1;

        for (index=0; index < gf_elements-1; index++){
                this->gflog[b] = (unsigned char)index;
                this->gfilog[index] = (unsigned char)b;
                b <<= 1;
                if (b & gf_elements){
                        b ^= prim_poly_8;
                }
        }

        return 0;
}


/*
 * gf_mul: Multiple a and b 
 *
 * @param a, b: input values for multiplication
 * @param s: field size (2 power s)
 *
 * @return: Product in finite field
 */

unsigned short CodingLayer::gf_mul(int a, int b, int s)
{
        unsigned int field_max;
        int sum;
        unsigned short result;

        if ((a == 0) || (b == 0)){
                result = 0;
        }
        else{
                field_max = (1 << s) - 1;
                sum = gflog[a] + gflog[b];
                sum = sum % field_max;

                result = gfilog[sum];
        }

        return(result);
}


/*
 * gf_div: Divide a by b
 *
 * @param a, b: input values for division
 * @param s: field size (2 power s)
 *
 * @return: Divident in finite field
 */

unsigned short CodingLayer::gf_div(int a, int b, int s)
{
        unsigned int field_max;
        int diff;
        unsigned short result;

	if (a == 0){
		result = 0;
	}
	else{
        	field_max = (1 << s) - 1;

        	diff = gflog[a] - gflog[b];

        	if (diff < 0){
                	diff = diff + field_max;
        	}

        	result = gfilog[diff];
	}

        return(result);
}


/*
 * gf_get_coefficient: Get the coefficient of the primitive polynomial
 * polynomial = x^8 + x^4 + x^3 + x^2 + 1
 *
 * @param value: the coefficient offset
 * @param s: field size (2 power s)
 *
 * @return: the coefficient
 */

int CodingLayer::gf_get_coefficient(int value, int s)
{
        int result;

	result = 1 << value;

        return result;
}


/*************************************************************************
 * MBR functions
 *************************************************************************/

/*
 * mbr_find_block_id: find block id for a block based on its disk id and block no
 *
 * @param disk_id: disk ID of the data block
 * @param block_no: block number of the data block
 * @param mbr_segment_size: the segment size for MBR coding
 *
 * @return: mbr block id
 */
int CodingLayer::mbr_find_block_id(int disk_id, int block_no, int mbr_segment_size)
{
        int mbr_block_id;
        int block_offset, mbr_offset;
        int counter, i;

        block_offset = block_no % mbr_segment_size;

        if (block_offset < disk_id){
                mbr_block_id = -1;
        }
        else{
                mbr_offset = block_offset - disk_id;
                mbr_block_id = 0;
                counter = mbr_segment_size;
                for (i=0; i<disk_id; i++){
                        mbr_block_id = mbr_block_id + counter;
                        counter--;
                }

                mbr_block_id = mbr_block_id + mbr_offset;
        }

        return mbr_block_id;
}

/*
 * mbr_find_dup_block_id: find block id for a duplicated block based on its disk id and block no
 *
 * @param disk_id: disk ID of the data block
 * @param block_no: block number of the data block
 * @param mbr_segment_size: the segment size for MBR coding
 *
 * @return: mbr block id
 */
int CodingLayer::mbr_find_dup_block_id(int disk_id, int block_no, int mbr_segment_size)
{
        int mbr_block_id;
        int block_offset, mbr_offset;
        int counter, i;

        block_offset = block_no % mbr_segment_size;

        if (block_offset >= disk_id){
                mbr_block_id = -1;
        }
        else{
                mbr_offset = disk_id - 1;
                mbr_block_id = 0;
                counter = mbr_segment_size;
                for (i=0; i<block_offset; i++){
                        mbr_block_id = mbr_block_id + counter;
                        counter--;
                        mbr_offset--;
                }

                mbr_block_id = mbr_block_id + mbr_offset;
        }

        return mbr_block_id;
}


/*
 * mbr_get_disk_id: get disk id for a block based on its MBR block ID
 *
 * @param mbr_block_id: mbr block ID of the data block
 * @param mbr_segment_size: the segment size for MBR coding
 *
 * @return: disk id
 */
int CodingLayer::mbr_get_disk_id(int mbr_block_id, int mbr_segment_size)
{
        int mbr_block_group, counter, temp;

        mbr_block_group = 0;
        counter = mbr_segment_size;
        temp = (mbr_block_id + 1) - counter;

        while (temp > 0){
                counter--;
                temp = temp - counter;
                (mbr_block_group)++;
        }

        return mbr_block_group;
}


/*
 * mbr_get_block_no: get block number for a block based on its disk id and MBR block number
 *
 * @param disk id: disk ID of the data block
 * @param mbr_block_id: mbr block ID of the data block
 * @param mbr_segment_size: the segment size for MBR coding
 *
 * @return: block number
 */
int CodingLayer::mbr_get_block_no(int disk_id, int mbr_block_id, int mbr_segment_size)
{
        int block_no;
        int mbr_offset, j;

        mbr_offset = mbr_block_id;
        for (j=0; j<disk_id; j++){
                mbr_offset = mbr_offset - (mbr_segment_size - j);
        }
        block_no = mbr_offset + disk_id;

        return block_no;
}


/*
 * mbr_get_dup_disk_id: get disk id for a duplicated block based on its MBR block ID
 *
 * @param mbr_block_id: mbr block ID of the data block
 * @param mbr_segment_size: the segment size for MBR coding
 *
 * @return: disk id
 */
int CodingLayer::mbr_get_dup_disk_id(int mbr_block_id, int mbr_segment_size)
{
        int dup_disk_id;
        int disk_id, mbr_offset;
        int j;

        disk_id = mbr_get_disk_id(mbr_block_id, mbr_segment_size);
        mbr_offset = mbr_block_id;
        for (j=0; j<disk_id; j++){
                mbr_offset = mbr_offset - (mbr_segment_size - j);
        }

        dup_disk_id = mbr_offset + 1 + disk_id;

        return dup_disk_id;
}


/*
 * mbr_get_dup_block_no: get block number for a duplicated block based on its disk id and MBR block number
 *
 * @param disk id: disk ID of the original data block
 * @param mbr_block_id: mbr block ID of the data block
 * @param mbr_segment_size: the segment size for MBR coding
 *
 * @return: block number
 */
int CodingLayer::mbr_get_dup_block_no(int disk_id, int mbr_block_id, int mbr_segment_size)
{
        int dup_block_no;

        dup_block_no = disk_id;

        return dup_block_no;
}


/*************************************************************************
 * MSR functions
 *************************************************************************/

/*
 * msr_gen_coding_matrix: generate coding matrix for msr coding
 *
 * Note: currently hardcode for (6,3,5) MSR code
 *
 * @param msr_n: codeword size
 * @param msr_k: data size
 * @param msr_m: number of native blocks
 * @param msr_c: number of coded blocks
 *
 * @return: 0
 */
int CodingLayer::msr_gen_coding_matrix(int msr_n, int msr_k, int msr_m, int msr_c)
{
	int i;

	this->msr_coding_matrix = (unsigned short**)malloc(msr_c * sizeof(unsigned short*));
	for (i=0; i<msr_c; i++){
		msr_coding_matrix[i] = (unsigned short*)malloc(msr_m * sizeof(unsigned short));
	}


        //M'0 coding coefficients
        msr_coding_matrix[0][0] = 3;
        msr_coding_matrix[0][1] = 0;
        msr_coding_matrix[0][2] = 0;
        msr_coding_matrix[0][3] = 3;
        msr_coding_matrix[0][4] = 0;
        msr_coding_matrix[0][5] = 0;
        msr_coding_matrix[0][6] = 3;
        msr_coding_matrix[0][7] = 0;
        msr_coding_matrix[0][8] = 0;

        //M'1 coding coefficients
        msr_coding_matrix[1][0] = 2;
        msr_coding_matrix[1][1] = 1;
        msr_coding_matrix[1][2] = 0;
        msr_coding_matrix[1][3] = 3;
        msr_coding_matrix[1][4] = 1;
        msr_coding_matrix[1][5] = 0;
        msr_coding_matrix[1][6] = 1;
        msr_coding_matrix[1][7] = 1;
        msr_coding_matrix[1][8] = 0;

        //M'2 coding coefficients
        msr_coding_matrix[2][0] = 2;
        msr_coding_matrix[2][1] = 0;
        msr_coding_matrix[2][2] = 1;
        msr_coding_matrix[2][3] = 1;
        msr_coding_matrix[2][4] = 0;
        msr_coding_matrix[2][5] = 1;
        msr_coding_matrix[2][6] = 3;
        msr_coding_matrix[2][7] = 0;
        msr_coding_matrix[2][8] = 1;

        //M'3 coding coefficients
        msr_coding_matrix[3][0] = 1;
        msr_coding_matrix[3][1] = 2;
        msr_coding_matrix[3][2] = 0;
        msr_coding_matrix[3][3] = 2;
        msr_coding_matrix[3][4] = 2;
        msr_coding_matrix[3][5] = 0;
        msr_coding_matrix[3][6] = 3;
        msr_coding_matrix[3][7] = 2;
        msr_coding_matrix[3][8] = 0;

        //M'4 coding coefficients
        msr_coding_matrix[4][0] = 0;
        msr_coding_matrix[4][1] = 3;
        msr_coding_matrix[4][2] = 0;
        msr_coding_matrix[4][3] = 0;
        msr_coding_matrix[4][4] = 1;
        msr_coding_matrix[4][5] = 0;
        msr_coding_matrix[4][6] = 0;
        msr_coding_matrix[4][7] = 2;
        msr_coding_matrix[4][8] = 0;

        //M'5 coding coefficients
        msr_coding_matrix[5][0] = 0;
        msr_coding_matrix[5][1] = 2;
        msr_coding_matrix[5][2] = 1;
        msr_coding_matrix[5][3] = 0;
        msr_coding_matrix[5][4] = 1;
        msr_coding_matrix[5][5] = 2;
        msr_coding_matrix[5][6] = 0;
        msr_coding_matrix[5][7] = 3;
        msr_coding_matrix[5][8] = 3;

        //M'6 coding coefficients
        msr_coding_matrix[6][0] = 1;
        msr_coding_matrix[6][1] = 0;
        msr_coding_matrix[6][2] = 2;
        msr_coding_matrix[6][3] = 3;
        msr_coding_matrix[6][4] = 0;
        msr_coding_matrix[6][5] = 2;
        msr_coding_matrix[6][6] = 2;
        msr_coding_matrix[6][7] = 0;
        msr_coding_matrix[6][8] = 2;

        //M'7 coding coefficients
        msr_coding_matrix[7][0] = 0;
        msr_coding_matrix[7][1] = 1;
        msr_coding_matrix[7][2] = 2;
        msr_coding_matrix[7][3] = 0;
        msr_coding_matrix[7][4] = 3;
        msr_coding_matrix[7][5] = 3;
        msr_coding_matrix[7][6] = 0;
        msr_coding_matrix[7][7] = 2;
        msr_coding_matrix[7][8] = 1;

        //M'8 coding coefficients
        msr_coding_matrix[8][0] = 0;
        msr_coding_matrix[8][1] = 0;
        msr_coding_matrix[8][2] = 3;
        msr_coding_matrix[8][3] = 0;
        msr_coding_matrix[8][4] = 0;
        msr_coding_matrix[8][5] = 2;
        msr_coding_matrix[8][6] = 0;
        msr_coding_matrix[8][7] = 0;
        msr_coding_matrix[8][8] = 1;

        return(0);
}


/*
 * msr_find_block_id: find block id for a block based on its disk id and block no
 *
 * @param disk_id: disk ID of the data block
 * @param block_no: block number of the data block
 * @param msr_segment_size: the segment size for MSR coding
 * @param msr_n: codeword size
 * @param msr_k: data size
 *
 * @return: msr block id
 */
int CodingLayer::msr_find_block_id(int disk_id, int block_no, int msr_segment_size, int msr_n, int msr_k)
{
        int msr_block_id;
        int block_offset;

        block_offset = block_no % msr_segment_size;

        if (disk_id >= msr_k){
                msr_block_id = -1;
        }
        else{
                msr_block_id = disk_id * msr_segment_size + block_offset;
        }

        return msr_block_id;
}


/*
 * msr_find_code_block_id: find coded block id for a block based on its disk id and block no
 *
 * @param disk_id: disk ID of the data block
 * @param block_no: block number of the data block
 * @param msr_segment_size: the segment size for MSR coding
 * @param msr_n: codeword size
 * @param msr_k: data size
 *
 * @return: msr block id
 */
int CodingLayer::msr_find_code_block_id(int disk_id, int block_no, int msr_segment_size, int msr_n, int msr_k)
{
        int msr_block_id;
        int block_offset;

        block_offset = block_no % msr_segment_size;

        if (disk_id < msr_k){
                msr_block_id = -1;
        }
        else{
                msr_block_id = (disk_id - msr_k) * msr_segment_size + block_offset;
        }

        return msr_block_id;
}



//MSR get disk id for a block based on its MSR block number
/*
 * msr_get_disk_id: get disk id for a block based on its MSR block ID
 *
 * @param msr_block_id: msr block ID of the data block
 * @param msr_segment_size: the segment size for MSR coding
 * @param msr_n: codeword size
 * @param msr_k: data size
 *
 * @return: disk id
 */
int CodingLayer::msr_get_disk_id(int msr_block_id, int msr_segment_size, int msr_n, int msr_k)
{
        int disk_id;

        disk_id = (int)(msr_block_id / msr_segment_size);

        return disk_id;
}


/*
 * msr_get_block_no: get block number for a block based on its MSR block number
 *
 * @param msr_block_id: msr block ID of the data block
 * @param msr_segment_size: the segment size for MSR coding
 *
 * @return: block number
 */
int CodingLayer::msr_get_block_no(int msr_block_id, int msr_segment_size)
{
        int block_no;

        block_no = msr_block_id % msr_segment_size;

        return block_no;
}


/*
 * msr_get_code_disk_id: get coded disk id for a block based on its MSR block ID
 *
 * @param msr_block_id: msr block ID of the data block
 * @param msr_segment_size: the segment size for MSR coding
 * @param msr_n: codeword size
 * @param msr_k: data size
 *
 * @return: disk id
 */
int CodingLayer::msr_get_code_disk_id(int msr_block_id, int msr_segment_size, int msr_n, int msr_k)
{
        int disk_id;

        disk_id = (int)(msr_block_id / msr_segment_size);
        disk_id = disk_id + msr_k;

        return disk_id;
}


/*
 * msr_get_code_block_no: get coded block number for a block based on its MSR block number
 *
 * @param msr_block_id: msr block ID of the data block
 * @param msr_segment_size: the segment size for MSR coding
 *
 * @return: block number
 */
int CodingLayer::msr_get_code_block_no(int msr_block_id, int msr_segment_size)
{
        int block_no;

        block_no = msr_block_id % msr_segment_size;

        return block_no;
}


/*
 * msr_find_single_decode_block_id: get the decode block id for single node recovery
 *
 * @param msr_block_id: msr block ID of the data block
 * @param msr_segment_size: the segment size for MSR coding
 *
 * @return: block number
 */
int CodingLayer::msr_find_single_decode_block_id(int msr_block_id, int msr_segment_size)
{
        int decode_block_id;

        decode_block_id = (msr_block_id % msr_segment_size) * (msr_segment_size + 1);

        return decode_block_id;
}



/*************************************************************************
 * Encoding functions
 *************************************************************************/

/*
 * default_encoding: select first available block
 *
 * @param buf: buffer of data to be written to disk
 * @param size: size of the data buffer
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encoding_default(const char* buf, int size)
{

	int retstat, disk_id, block_no, disk_total_num, block_size;
	//char *dev;
	//int datafd;
	int size_request, block_request, free_size;
	struct data_block_info block_written;
	int i;

	//test print
	//printf("***get_data_block_no 1: size=%d\n",size);

	disk_total_num = NCFS_DATA->disk_total_num;
	block_size = NCFS_DATA->disk_block_size;

	size_request = fileSystemLayer->round_to_block_size(size);
	block_request = size_request / block_size;

	//test print
	//printf("***get_data_block_no 2: size_request=%d, block_request=%d\n"
	//		,size_request,block_request);


	//implement disk write algorithm here.
	//now use free block in bigger free space.

	block_no = 0;
	disk_id = -1;
	free_size = 0;
	for (i=0; i < disk_total_num; i++){
		if ((block_request <= (NCFS_DATA->free_size[i]))
				&& (free_size < (NCFS_DATA->free_size[i])))
		{
			disk_id = i;
			block_no = NCFS_DATA->free_offset[i];
			free_size = NCFS_DATA->free_size[i];
		}
	}


	//get block from space_list if no free block available
	if (disk_id == -1){
		if (NCFS_DATA->space_list_head != NULL){
			disk_id = NCFS_DATA->space_list_head->disk_id;
			block_no = NCFS_DATA->space_list_head->disk_block_no;
			fileSystemLayer->space_list_remove(disk_id,block_no);
		}
	}

	if (disk_id == -1){	//if no free block and no space finally
		printf("***get_data_block_no: ERROR disk_id = -1\n");
	}
	else{
		//Cache Start
		retstat = cacheLayer->DiskWrite(disk_id,buf,size,block_no*block_size);
		//Cache End

		NCFS_DATA->free_offset[disk_id] = block_no + block_request;
		NCFS_DATA->free_size[disk_id]
			= NCFS_DATA->free_size[disk_id] - block_request;

	}

	block_written.disk_id = disk_id;
	block_written.block_no = block_no;

	return block_written;

}


/*
 * jbod_encoding: JBOD: non-stripped block allocation  (type=100)
 *
 * @param buf: buffer of data to be written to disk
 * @param size: size of the data buffer
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encoding_jbod(const char* buf, int size)
{
	int retstat, disk_id, block_no, disk_total_num, block_size;
	//char *dev;
	//int datafd;
	int size_request, block_request;
	struct data_block_info block_written;
	int i;

	//test print
	//printf("***get_data_block_no 1: size=%d\n",size);

	disk_total_num = NCFS_DATA->disk_total_num;
	block_size = NCFS_DATA->disk_block_size;

	size_request = fileSystemLayer->round_to_block_size(size);
	block_request = size_request / block_size;

	//test print
	//printf("***get_data_block_no 2: size_request=%d, block_request=%d\n"
	//		,size_request,block_request);


	//implement disk write algorithm here.
	//here use jbod: first available block.

	block_no = 0;
	disk_id = -1;
	i = 0;
	while ((i < disk_total_num) && (disk_id == -1)){
		if (block_request <= (NCFS_DATA->free_size[i]))
		{
			disk_id = i;
			block_no = NCFS_DATA->free_offset[i];
		}
		i++;
	}


	//get block from space_list if no free block available
	if (disk_id == -1){
		if (NCFS_DATA->space_list_head != NULL){
			disk_id = NCFS_DATA->space_list_head->disk_id;
			block_no = NCFS_DATA->space_list_head->disk_block_no;
			fileSystemLayer->space_list_remove(disk_id,block_no);
		}
	}


	if (disk_id == -1){
		printf("***get_data_block_no: ERROR disk_id = -1\n");
	}
	else{
		//Cache Start

		retstat = cacheLayer->DiskWrite(disk_id,buf,size,block_no*block_size);

		NCFS_DATA->free_offset[disk_id] = block_no + block_request;
		NCFS_DATA->free_size[disk_id]
			= NCFS_DATA->free_size[disk_id] - block_request;

		//Cache End
	}

	block_written.disk_id = disk_id;
	block_written.block_no = block_no;

	return block_written;

}


/*
 * raid0_encoding: RAID 0: stripped block allocation (type=0)
 *
 * @param buf: buffer of data to be written to disk
 * @param size: size of the data buffer
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encoding_raid0(const char* buf, int size)
{
	int retstat, disk_id, block_no, disk_total_num, block_size;
	//char *dev;
	//int datafd;
	int size_request, block_request, free_offset;
	struct data_block_info block_written;
	int i;

	//test print
	//printf("***get_data_block_no 1: size=%d\n",size);

	disk_total_num = NCFS_DATA->disk_total_num;
	block_size = NCFS_DATA->disk_block_size;

	size_request = fileSystemLayer->round_to_block_size(size);
	block_request = size_request / block_size;

	//test print
	//printf("***get_data_block_no 2: size_request=%d, block_request=%d\n"
	//		,size_request,block_request);


	//implement disk write algorithm here.
	//here use raid0: stripped block allocation.

	block_no = 0;
	disk_id = -1;
	free_offset = NCFS_DATA->free_offset[0];
	for (i=disk_total_num-1; i >= 0; i--){
		if ((block_request <= (NCFS_DATA->free_size[i]))
				&& (free_offset >= (NCFS_DATA->free_offset[i])))
		{
			disk_id = i;
			block_no = NCFS_DATA->free_offset[i];
			free_offset = block_no;
		}
	}


	//get block from space_list if no free block available
	if (disk_id == -1){
		//printf("***raid0: get_space_list: list_num=%d\n",NCFS_DATA->space_list_num);
		if (NCFS_DATA->space_list_head != NULL){
			//printf("***raid0: get_space_list: list_head != NULL \n");
			disk_id = NCFS_DATA->space_list_head->disk_id;
			block_no = NCFS_DATA->space_list_head->disk_block_no;
			fileSystemLayer->space_list_remove(disk_id,block_no);
		}
	}


	if (disk_id == -1){
		printf("***get_data_block_no: ERROR disk_id = -1\n");
	}
	else{
		//Cache Start
		retstat = cacheLayer->DiskWrite(disk_id,buf,size,block_no*block_size);

		NCFS_DATA->free_offset[disk_id] = block_no + block_request;
		NCFS_DATA->free_size[disk_id]
			= NCFS_DATA->free_size[disk_id] - block_request;
		//Cache End
	}

	block_written.disk_id = disk_id;
	block_written.block_no = block_no;

	return block_written;

}


/*
 * raid1_encoding: RAID 1: fault tolerance by mirror (type=1)
 *
 * @param buf: buffer of data to be written to disk
 * @param size: size of the data buffer
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encoding_raid1(const char* buf, int size)
{
	int retstat, disk_id, block_no, disk_total_num, block_size;
	//char *dev, *dev2;
	//int datafd, datafd2;
	int size_request, block_request;
	struct data_block_info block_written;
	int i;
	int mirror_disk_id;

	//test print
	//printf("***get_data_block_no 1: size=%d\n",size);

	disk_total_num = NCFS_DATA->disk_total_num;
	block_size = NCFS_DATA->disk_block_size;

	size_request = fileSystemLayer->round_to_block_size(size);
	block_request = size_request / block_size;

	//test print
	//printf("***get_data_block_no 2: size_request=%d, block_request=%d\n"
	//		,size_request,block_request);


	//implement disk write algorithm here.
	//here use raid1: mirroring. (with jbod)

	block_no = 0;
	disk_id = -1;
	i = 0;
	//first n/2 disks are data disks
	while ((i < disk_total_num/2) && (disk_id == -1)){
		if (block_request <= (NCFS_DATA->free_size[i]))
		{
			disk_id = i;
			block_no = NCFS_DATA->free_offset[i];
		}
		i++;
	}


	//get block from space_list if no free block available
	if (disk_id == -1){
		if (NCFS_DATA->space_list_head != NULL){
			disk_id = NCFS_DATA->space_list_head->disk_id;
			block_no = NCFS_DATA->space_list_head->disk_block_no;
			fileSystemLayer->space_list_remove(disk_id,block_no);
		}
	}


	if (disk_id == -1){
		printf("***get_data_block_no: ERROR disk_id = -1\n");
	}
	else{
		//Cache Start
		retstat = cacheLayer->DiskWrite(disk_id,buf,size,block_no*block_size);
		//Cache End

		NCFS_DATA->free_offset[disk_id] = block_no + block_request;
		NCFS_DATA->free_size[disk_id]
			= NCFS_DATA->free_size[disk_id] - block_request;

		//write data to mirror disk
		mirror_disk_id = disk_total_num - disk_id - 1;
		//Cache Start
		retstat = cacheLayer->DiskWrite(mirror_disk_id,buf,size,block_no*block_size);
		//Cache End
	}

	block_written.disk_id = disk_id;
	block_written.block_no = block_no;

	return block_written;
}


/*
 * raid4_encoding: RAID 4: fault tolerance by non-stripped parity (type=4)
 *
 * @param buf: buffer of data to be written to disk
 * @param size: size of the data buffer
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encoding_raid4(const char* buf, int size)
{
	int retstat, disk_id, block_no, disk_total_num, block_size;
	//char *dev, *dev2;
	//int datafd, datafd2;
	int size_request, block_request, free_offset;
	struct data_block_info block_written;
	int i, j;
	int parity_disk_id;
	char *buf2, *buf_read;

	//test print
	//printf("***get_data_block_no 1: size=%d\n",size);

	disk_total_num = NCFS_DATA->disk_total_num;
	block_size = NCFS_DATA->disk_block_size;

	size_request = fileSystemLayer->round_to_block_size(size);
	block_request = size_request / block_size;

	//test print
	//printf("***get_data_block_no 2: size_request=%d, block_request=%d\n"
	//		,size_request,block_request);


	//implement disk write algorithm here.
	//here use raid4: stripped block allocation plus dedicated parity disk.
	//use the last disk as parity disk.
	//approach: calculate xor of the block on all data disk 
	block_no = 0;
	disk_id = -1;
	free_offset = NCFS_DATA->free_offset[0];
	for (i=disk_total_num-2; i >= 0; i--){
		if ((block_request <= (NCFS_DATA->free_size[i]))
				&& (free_offset >= (NCFS_DATA->free_offset[i])))
		{
			disk_id = i;
			block_no = NCFS_DATA->free_offset[i];
			free_offset = block_no;
		}
	}


	//get block from space_list if no free block available
	if (disk_id == -1){
		if (NCFS_DATA->space_list_head != NULL){
			disk_id = NCFS_DATA->space_list_head->disk_id;
			block_no = NCFS_DATA->space_list_head->disk_block_no;
			fileSystemLayer->space_list_remove(disk_id,block_no);
		}
	}


	if (disk_id == -1){
		printf("***get_data_block_no: ERROR disk_id = -1\n");
	}
	else{
		//start ncfs 0.11 michael. Use improved method for parity.
		buf_read = (char*)malloc(sizeof(char)*size_request);
		buf2 = (char*)malloc(sizeof(char)*size_request);
		//Cache Start
		retstat = cacheLayer->DiskRead(disk_id,buf2,size_request,block_no*block_size);
		//Cache End
		for (j=0; j < size_request; j++){
			buf2[j] = buf2[j] ^ buf[j];
		}
		retstat = cacheLayer->DiskWrite(disk_id,buf,size,block_no*block_size);

		NCFS_DATA->free_offset[disk_id] = block_no + block_request;
		NCFS_DATA->free_size[disk_id]
			= NCFS_DATA->free_size[disk_id] - block_request;


		parity_disk_id = disk_total_num - 1;
		retstat = cacheLayer->DiskRead(parity_disk_id,buf_read,size_request,block_no*block_size);
		for (j=0; j < size_request; j++){
			buf2[j] = buf2[j] ^ buf_read[j];
		}

		retstat = cacheLayer->DiskWrite(parity_disk_id,buf2,size,block_no*block_size);
		//printf("***raid4: retstat=%d\n",retstat);

		free(buf_read);
		free(buf2);
		//end ncfs 0.11 michael
	}

	block_written.disk_id = disk_id;
	block_written.block_no = block_no;

	return block_written;
}


/*
 * raid5_encoding: RAID 5: fault tolerance by stripped parity (type=5)
 * 
 * @param buf: buffer of data to be written to disk
 * @param size: size of the data buffer
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encoding_raid5(const char* buf, int size)
{
	int retstat, disk_id, block_no, disk_total_num, block_size;
	//char *dev, *dev2;
	//int datafd, datafd2;
	int size_request, block_request, free_offset;
	struct data_block_info block_written;
	int i, j;
	int parity_disk_id;
	char *buf2, *buf_read;

	struct timeval t1, t2, t3, t4, t5, t6;
	double duration;

	//test print
	//printf("***get_data_block_no 1: size=%d\n",size);

	disk_total_num = NCFS_DATA->disk_total_num;
	block_size = NCFS_DATA->disk_block_size;

	size_request = fileSystemLayer->round_to_block_size(size);
	block_request = size_request / block_size;

	//test print
	//printf("***get_data_block_no 2: size_request=%d, block_request=%d\n"
	//		,size_request,block_request);


	//implement disk write algorithm here.
	//here use raid5: stripped block allocation plus distributed parity.
	//approach: calculate xor of the block on all data disk
	for (i=0; i < disk_total_num; i++){
		if (i ==
				(disk_total_num-1-(NCFS_DATA->free_offset[i] % disk_total_num))){
			(NCFS_DATA->free_offset[i])++;
			(NCFS_DATA->free_size[i])--;
		}
	}

	block_no = 0;
	disk_id = -1;
	free_offset = NCFS_DATA->free_offset[0];
	for (i=disk_total_num-1; i >= 0; i--){
		if ((block_request <= (NCFS_DATA->free_size[i]))
				&& (free_offset >= (NCFS_DATA->free_offset[i])))
		{
			disk_id = i;
			block_no = NCFS_DATA->free_offset[i];
			free_offset = block_no;
		}
	}


	//get block from space_list if no free block available
	if (disk_id == -1){
		if (NCFS_DATA->space_list_head != NULL){
			disk_id = NCFS_DATA->space_list_head->disk_id;
			block_no = NCFS_DATA->space_list_head->disk_block_no;
			fileSystemLayer->space_list_remove(disk_id,block_no);
		}
	}


	if (disk_id == -1){
		printf("***get_data_block_no: ERROR disk_id = -1\n");
	}
	else{
		//start ncfs 0.11 michael. Use improved method for parity.
		buf_read = (char*)malloc(sizeof(char)*size_request);
		buf2 = (char*)malloc(sizeof(char)*size_request);

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t1,NULL);
                }

		//Cache Start
		retstat = cacheLayer->DiskRead(disk_id,buf2,size_request,block_no*block_size);
		//Cache End

		if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t2,NULL);
          	}


		for (j=0; j < size_request; j++){
			buf2[j] = buf2[j] ^ buf[j];
		}

		NCFS_DATA->free_offset[disk_id] = block_no + block_request;
		NCFS_DATA->free_size[disk_id]
			= NCFS_DATA->free_size[disk_id] - block_request;


		parity_disk_id = disk_total_num - 1 - (block_no % disk_total_num);

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t3,NULL);
                }

		//Cache Start
		retstat = cacheLayer->DiskRead(parity_disk_id,buf_read,size_request,block_no*block_size);
		//Cache End

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t4,NULL);
                }


		for (j=0; j < size_request; j++){
			buf2[j] = buf2[j] ^ buf_read[j];
		}

		if (NCFS_DATA->run_experiment == 1){
			gettimeofday(&t5,NULL);
		}

                //Cache Start
		//write data
                retstat = cacheLayer->DiskWrite(disk_id,buf,size,block_no*block_size);
                //Cache End


		//Cache Start
		//write parity
		retstat = cacheLayer->DiskWrite(parity_disk_id,buf2,size,block_no*block_size);
		//Cache End
		//printf("***raid5: retstat=%d\n",retstat);

		free(buf_read);
		free(buf2);
		//end ncfs 0.11 michael

		if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t6,NULL);

                        //duration = (t2.tv_sec - t1.tv_sec) + 
                        //        (t2.tv_usec-t1.tv_usec)/1000000.0;
                        //NCFS_DATA->diskread_time += duration;

                	duration = (t3.tv_sec - t2.tv_sec) + 
                                (t3.tv_usec-t2.tv_usec)/1000000.0;
                	NCFS_DATA->encoding_time += duration;

                        //duration = (t4.tv_sec - t3.tv_sec) + 
                        //        (t4.tv_usec-t3.tv_usec)/1000000.0;
                        //NCFS_DATA->diskread_time += duration;

                        duration = (t5.tv_sec - t4.tv_sec) + 
                                (t5.tv_usec-t4.tv_usec)/1000000.0;
                        NCFS_DATA->encoding_time += duration;

                        //duration = (t6.tv_sec - t5.tv_sec) + 
                        //        (t6.tv_usec-t5.tv_usec)/1000000.0;
                        //NCFS_DATA->diskwrite_time += duration;
		}
	}

	block_written.disk_id = disk_id;
	block_written.block_no = block_no;

	return block_written;
}



/*
 * raid6_encoding: RAID 6: fault tolerance by stripped parity (type=6)
 *
 * @param buf: buffer of data to be written to disk
 * @param size: size of the data buffer
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encoding_raid6(const char* buf, int size)
{
        int retstat, disk_id, block_no, disk_total_num, block_size;
        int size_request, block_request, free_offset;
        struct data_block_info block_written;
        int i, j;
        int parity_disk_id, code_disk_id;
        char *buf2, *buf3, *buf_read;
        char temp_char;
	int data_disk_coeff;

        struct timeval t1, t2, t3;
        double duration;

        //test print
        //printf("***get_data_block_no 1: size=%d\n",size);

        disk_total_num = NCFS_DATA->disk_total_num;
        block_size = NCFS_DATA->disk_block_size;

        size_request = fileSystemLayer->round_to_block_size(size);
        block_request = size_request / block_size;

        //test print
        //printf("***get_data_block_no 2: size_request=%d, block_request=%d\n", size_request, block_request);

        //implement disk write algorithm here.
        //here use raid6: stripped block allocation plus distributed parity and distributed RS code.


        for (i=0; i < disk_total_num; i++){
                //mark blocks of code_disk and parity_disk
                if ((i == (disk_total_num-1-(NCFS_DATA->free_offset[i] % disk_total_num)))
                || (i == (disk_total_num-1-((NCFS_DATA->free_offset[i]+1) % disk_total_num))))
                {
                        (NCFS_DATA->free_offset[i])++;
                        (NCFS_DATA->free_size[i])--;
                }
        }


        block_no = 0;
        disk_id = -1;
        free_offset = NCFS_DATA->free_offset[0];
        for (i=disk_total_num-1; i >= 0; i--){
//        for (i=disk_total_num-3; i>=0; i--){
                if ((block_request <= (NCFS_DATA->free_size[i]))
                                && (free_offset >= (NCFS_DATA->free_offset[i])))
                {
                        disk_id = i;
                        block_no = NCFS_DATA->free_offset[i];
                        free_offset = block_no;
                }
        }


        //get block from space_list if no free block available
        if (disk_id == -1){
                if (NCFS_DATA->space_list_head != NULL){
                        disk_id = NCFS_DATA->space_list_head->disk_id;
                        block_no = NCFS_DATA->space_list_head->disk_block_no;
                        fileSystemLayer->space_list_remove(disk_id,block_no);
                }
        }


        if (disk_id == -1){
                printf("***get_data_block_no: ERROR disk_id = -1\n");
        }
        else{
                buf_read = (char*)malloc(sizeof(char)*size_request);
                buf2 = (char*)malloc(sizeof(char)*size_request);
                memset(buf2, 0, size_request);
                buf3 = (char*)malloc(sizeof(char)*size_request);
                memset(buf3, 0, size_request);

                NCFS_DATA->free_offset[disk_id] = block_no + block_request;
                NCFS_DATA->free_size[disk_id]
                        = NCFS_DATA->free_size[disk_id] - block_request;

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t1,NULL);
                }

                // Cache Start
                retstat = cacheLayer->DiskWrite(disk_id,buf,size,block_no*block_size);
		// Cache End

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t2,NULL);

                        //duration = (t2.tv_sec - t1.tv_sec) + 
                        //        (t2.tv_usec-t1.tv_usec)/1000000.0;
                        //NCFS_DATA->diskwrite_time += duration;
                }

                code_disk_id = disk_total_num -1 - (block_no % disk_total_num);
                parity_disk_id = disk_total_num - 1 - ((block_no+1) % disk_total_num);
                //code_disk_id = disk_total_num - 1;
                //parity_disk_id = disk_total_num - 2;

                for (i=0; i < disk_total_num; i++){
                        if ((i != parity_disk_id) && (i != code_disk_id)){

				if (NCFS_DATA->run_experiment == 1){
                        		gettimeofday(&t1,NULL);
                		}

                                //Cache Start
                                retstat = cacheLayer->DiskRead(i,buf_read,
                                                        size_request,block_no*block_size);
                                //Cache End

                                if (NCFS_DATA->run_experiment == 1){
                                        gettimeofday(&t2,NULL);
                                }

                                for (j=0; j < size_request; j++){
                                        //Calculate parity block P
                                        buf2[j] = buf2[j] ^ buf_read[j];

                                	//calculate the coefficient of the data block
                                	data_disk_coeff = i;

                                	if (i > code_disk_id){
                                        	(data_disk_coeff)--;
                                	}
                                	if (i > parity_disk_id){
                                        	(data_disk_coeff)--;
                                	}
                                	data_disk_coeff = disk_total_num - 3 - data_disk_coeff;
                                	data_disk_coeff = gf_get_coefficient(data_disk_coeff, field_power);

                                	//calculate code block Q
                                	temp_char = buf3[j];
                                	buf3[j] = temp_char ^
                                        (char)gf_mul((unsigned char)buf_read[j],data_disk_coeff,field_power);
                                }

                                if (NCFS_DATA->run_experiment == 1){
                                        gettimeofday(&t3,NULL);

                        		//duration = (t2.tv_sec - t1.tv_sec) + 
                                	//	(t2.tv_usec-t1.tv_usec)/1000000.0;
                        		//NCFS_DATA->diskread_time += duration;

                        		duration = (t3.tv_sec - t2.tv_sec) + 
                                		(t3.tv_usec-t2.tv_usec)/1000000.0;
                        		NCFS_DATA->encoding_time += duration;
                                }
                        }
                }

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t1,NULL);
                }

                // Cache Start
                retstat = cacheLayer->DiskWrite(parity_disk_id,buf2,size,block_no*block_size);
                retstat = cacheLayer->DiskWrite(code_disk_id,buf3,size,block_no*block_size);
                // Cache End
                //printf("***raid6: retstat=%d\n",retstat);

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t2,NULL);

                        //duration = (t2.tv_sec - t1.tv_sec) + 
                        //        (t2.tv_usec-t1.tv_usec)/1000000.0;
                        //NCFS_DATA->diskwrite_time += duration;
                }

                free(buf_read);
                free(buf2);
                free(buf3);
        }

        block_written.disk_id = disk_id;
        block_written.block_no = block_no;

        return block_written;
}


/*
 * mbr_encoding: MBR: Minimum Bandwidth Regenerating code (type=1000)
 *
 * @param buf: buffer of data to be written to disk
 * @param size: size of the data buffer
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encoding_mbr(const char* buf, int size)
{
        int retstat, disk_id, block_no, disk_total_num, block_size;
        int size_request, block_request, free_offset;
        struct data_block_info block_written;
        int i, j;
        char *buf2, *buf_read;

	int mbr_segment_size;
	int mbr_segment_id;
	int mbr_block_id;
	int dup_disk_id, dup_block_no, code_disk_id, code_block_no;

        struct timeval t1, t2, t3, t4;
        double duration;

	mbr_segment_size = NCFS_DATA->mbr_segment_size;

        //test print
        //printf("***get_data_block_no 1: size=%d\n",size);

        disk_total_num = NCFS_DATA->disk_total_num;
        block_size = NCFS_DATA->disk_block_size;

        size_request = fileSystemLayer->round_to_block_size(size);
        block_request = size_request / block_size;

        //test print
        //printf("***get_data_block_no 2: size_request=%d, block_request=%d\n"
        //                ,size_request,block_request);

        //implement disk write algorithm here.
        //here use mbr.
        //approach: write original data block, duplicated block and coded block
        for (i=0; i < disk_total_num; i++){
		mbr_block_id = mbr_find_block_id(i,NCFS_DATA->free_offset[i],mbr_segment_size);
                if ( (mbr_block_id == -1) || (mbr_block_id >= NCFS_DATA->mbr_m) ){
                        (NCFS_DATA->free_offset[i])++;
                        (NCFS_DATA->free_size[i])--;
                }
        }

        block_no = 0;
        disk_id = -1;
        free_offset = NCFS_DATA->free_offset[0];
        for (i=disk_total_num-1; i >= 0; i--){
                mbr_block_id = mbr_find_block_id(i,NCFS_DATA->free_offset[i],mbr_segment_size);
                if ((block_request <= (NCFS_DATA->free_size[i]))
                                && (free_offset >= (NCFS_DATA->free_offset[i]))
				&& (mbr_block_id != -1)
				&& (mbr_block_id < NCFS_DATA->mbr_m))
                {
                        disk_id = i;
                        block_no = NCFS_DATA->free_offset[i];
                        free_offset = block_no;
                }
        }


        //get block from space_list if no free block available
        if (disk_id == -1){
                if (NCFS_DATA->space_list_head != NULL){
                        disk_id = NCFS_DATA->space_list_head->disk_id;
                        block_no = NCFS_DATA->space_list_head->disk_block_no;
                        fileSystemLayer->space_list_remove(disk_id,block_no);
                }
        }


        if (disk_id == -1){
                printf("***get_data_block_no: ERROR disk_id = -1\n");
        }
        else{
		mbr_segment_id = (int)(block_no / mbr_segment_size);

		//write original data block
                buf_read = (char*)malloc(sizeof(char)*size_request);
                buf2 = (char*)malloc(sizeof(char)*size_request);


		if (NCFS_DATA->mbr_c > 0){

                	if (NCFS_DATA->run_experiment == 1){
                        	gettimeofday(&t1,NULL);
                	}

                	//Cache Start
                	retstat = cacheLayer->DiskRead(disk_id,buf2,size_request,block_no*block_size);
                	//Cache End

                	if (NCFS_DATA->run_experiment == 1){
                        	gettimeofday(&t2,NULL);
                	}

                	for (j=0; j < size_request; j++){
                        	buf2[j] = buf2[j] ^ buf[j];
                	}

                	if (NCFS_DATA->run_experiment == 1){
                        	gettimeofday(&t3,NULL);

                        	//duration = (t2.tv_sec - t1.tv_sec) + 
                                //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                        	//NCFS_DATA->diskread_time += duration;

                        	duration = (t3.tv_sec - t2.tv_sec) + 
                                	(t3.tv_usec-t2.tv_usec)/1000000.0;
                        	NCFS_DATA->encoding_time += duration;
                	}
		}

                NCFS_DATA->free_offset[disk_id] = block_no + block_request;
                NCFS_DATA->free_size[disk_id]
                        = NCFS_DATA->free_size[disk_id] - block_request;

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t1,NULL);
                }

		//write duplicated block
		mbr_block_id = mbr_find_block_id(disk_id, block_no, mbr_segment_size);
                dup_disk_id = mbr_get_dup_disk_id(mbr_block_id, mbr_segment_size);
                dup_block_no = mbr_get_dup_block_no(disk_id, mbr_block_id, mbr_segment_size);
                dup_block_no = dup_block_no + mbr_segment_id * mbr_segment_size;

		if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t2,NULL);
                }

                //Cache Start
		//write data
                retstat = cacheLayer->DiskWrite(disk_id,buf,size,block_no*block_size);
		//write duplicate data
                retstat = cacheLayer->DiskWrite(dup_disk_id,buf,size,dup_block_no*block_size);
                //Cache End

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t3,NULL);

                        duration = (t2.tv_sec - t1.tv_sec) + 
                                (t2.tv_usec-t1.tv_usec)/1000000.0;
                        NCFS_DATA->encoding_time += duration;

                        //duration = (t3.tv_sec - t2.tv_sec) + 
                        //        (t3.tv_usec-t2.tv_usec)/1000000.0;
                        //NCFS_DATA->diskwrite_time += duration;
                }


		if (NCFS_DATA->no_gui == 0){
			diskusageLayer->Update(dup_disk_id,dup_block_no,1);
		}


		if (NCFS_DATA->mbr_c > 0){

                        if (NCFS_DATA->run_experiment == 1){
                                gettimeofday(&t1,NULL);
                        }

                	//generate coded block data
                	mbr_block_id = NCFS_DATA->mbr_m; //use the first coded block; to be changed.
                	code_disk_id = mbr_get_disk_id(mbr_block_id, mbr_segment_size);
                	code_block_no = mbr_get_block_no(code_disk_id, mbr_block_id, mbr_segment_size);
                	code_block_no = code_block_no + mbr_segment_id * mbr_segment_size;

                	if (NCFS_DATA->run_experiment == 1){
                        	gettimeofday(&t2,NULL);
                	}

                	//Cache Start
                	retstat = cacheLayer->DiskRead(code_disk_id,buf_read,size_request,
						code_block_no*block_size);
                	//Cache End

			if (NCFS_DATA->run_experiment == 1){
                                gettimeofday(&t3,NULL);
                        }

                	for (j=0; j < size_request; j++){
                        	buf2[j] = buf2[j] ^ buf_read[j]; 
				//currently use XOR; to be changed later
                	}

                        if (NCFS_DATA->run_experiment == 1){
                                gettimeofday(&t4,NULL);

                        	duration = (t2.tv_sec - t1.tv_sec) + 
                                	(t2.tv_usec-t1.tv_usec)/1000000.0;
                        	NCFS_DATA->encoding_time += duration;

                        	//duration = (t3.tv_sec - t2.tv_sec) + 
                                //	(t3.tv_usec-t2.tv_usec)/1000000.0;
                        	//NCFS_DATA->diskread_time += duration;

                                duration = (t4.tv_sec - t3.tv_sec) + 
                                        (t4.tv_usec-t3.tv_usec)/1000000.0;
                                NCFS_DATA->encoding_time += duration;
                        }
		}


		for (j=0; j < NCFS_DATA->mbr_c; j++){

			if (NCFS_DATA->run_experiment == 1){
                        	gettimeofday(&t1,NULL);
                	}

			//write coded block
			mbr_block_id = (NCFS_DATA->mbr_m) + j;
			code_disk_id = mbr_get_disk_id(mbr_block_id, mbr_segment_size);
			code_block_no = mbr_get_block_no(code_disk_id, mbr_block_id, mbr_segment_size);
			code_block_no = code_block_no + mbr_segment_id * mbr_segment_size;

			//write duplicate coded block
                        dup_disk_id = mbr_get_dup_disk_id(mbr_block_id, mbr_segment_size);
                        dup_block_no = mbr_get_dup_block_no(code_disk_id, mbr_block_id, mbr_segment_size);
                        dup_block_no = dup_block_no + mbr_segment_id * mbr_segment_size;

			if (NCFS_DATA->run_experiment == 1){
                        	gettimeofday(&t2,NULL);
                	}

                        // Cache Start
			//write coded block
                        retstat = cacheLayer->DiskWrite(code_disk_id,buf2,size,code_block_no*block_size);
			//write duplicate coded block
                        retstat = cacheLayer->DiskWrite(dup_disk_id,buf2,size,dup_block_no*block_size);
                        // Cache End
                        //printf("***mbr: retstat2=%d\n",retstat);

			if (NCFS_DATA->run_experiment == 1){
                        	gettimeofday(&t3,NULL);

                        	duration = (t2.tv_sec - t1.tv_sec) + 
                                	(t2.tv_usec-t1.tv_usec)/1000000.0;
                        	NCFS_DATA->encoding_time += duration;

                        	//duration = (t3.tv_sec - t2.tv_sec) + 
                                //	(t3.tv_usec-t2.tv_usec)/1000000.0;
                        	//NCFS_DATA->diskwrite_time += duration;
                	}
		}

                free(buf_read);
                free(buf2);
	}

        block_written.disk_id = disk_id;
        block_written.block_no = block_no;

        return block_written;
}


/*
 * msr_encoding: MSR: Minimum Storage Regenerating code (type=2000)
 *
 * @param buf: buffer of data to be written to disk
 * @param size: size of the data buffer
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encoding_msr(const char* buf, int size)
{
        int retstat, disk_id, block_no, disk_total_num, block_size;
        int size_request, block_request, free_offset;
        struct data_block_info block_written;
        int i, j, k;
        char *buf2, *buf_read;

	int msr_n, msr_k, msr_m, msr_c;
        int msr_segment_size;
        int msr_segment_id;
        int msr_block_id;
        int code_disk_id, code_block_no;
	int coefficient;
	char encoded_data;

	msr_n = NCFS_DATA->msr_n;
	msr_k = NCFS_DATA->msr_k;
	msr_m = NCFS_DATA->msr_m;
	msr_c = NCFS_DATA->msr_c;
        msr_segment_size = NCFS_DATA->msr_segment_size;

        //test print
        //printf("***get_data_block_no 1: size=%d\n",size);

        disk_total_num = NCFS_DATA->disk_total_num;
        block_size = NCFS_DATA->disk_block_size;

        size_request = fileSystemLayer->round_to_block_size(size);
        block_request = size_request / block_size;

        //test print
        //printf("***get_data_block_no 2: size_request=%d, block_request=%d\n"
        //                ,size_request,block_request);

        //implement disk write algorithm here.
        //here use msr.
        //approach: write original data block, and coded block
        block_no = 0;
        disk_id = -1;
        free_offset = NCFS_DATA->free_offset[0];
        for (i=(int)(disk_total_num/2)-1; i >= 0; i--){
                if ( (block_request <= (NCFS_DATA->free_size[i]))
                                && (free_offset >= (NCFS_DATA->free_offset[i])) )
                {
                        disk_id = i;
                        block_no = NCFS_DATA->free_offset[i];
                        free_offset = block_no;
                }
        }

        //get block from space_list if no free block available
        if (disk_id == -1){
                if (NCFS_DATA->space_list_head != NULL){
                        disk_id = NCFS_DATA->space_list_head->disk_id;
                        block_no = NCFS_DATA->space_list_head->disk_block_no;
                        fileSystemLayer->space_list_remove(disk_id,block_no);
                }
        }


        if (disk_id == -1){
                printf("***get_data_block_no: ERROR disk_id = -1\n");
        }
        else{
                msr_segment_id = (int)(block_no / msr_segment_size);
		msr_block_id = msr_find_block_id(disk_id, block_no, msr_segment_size, msr_n, msr_k);

                //write original data block
                buf_read = (char*)malloc(sizeof(char)*size);
                buf2 = (char*)malloc(sizeof(char)*size);
                //Cache Start
                retstat = cacheLayer->DiskRead(disk_id,buf2,size,block_no*block_size);
                //Cache End

                //Cache Start
                retstat = cacheLayer->DiskWrite(disk_id,buf,size,block_no*block_size);
                //Cache End

                NCFS_DATA->free_offset[disk_id] = block_no + block_request;
                NCFS_DATA->free_size[disk_id]
                        = NCFS_DATA->free_size[disk_id] - block_request;

                for (j=0; j < msr_c; j++){
		coefficient = msr_coding_matrix[j][msr_block_id];
		if (coefficient != 0){

                        //get disk id an block no of coded block
                        code_disk_id = msr_get_code_disk_id(j, msr_segment_size, msr_n, msr_k);
                        code_block_no = msr_get_code_block_no(j, msr_segment_size);
                        code_block_no = code_block_no + msr_segment_id * msr_segment_size;

                	//Cache Start
                	retstat = cacheLayer->DiskRead(code_disk_id,buf_read,size,code_block_no*block_size);
                	//Cache End

			//generate coded block data
			for (k=0; k<size; k++){
				encoded_data = (char)gf_mul((unsigned char)buf2[k],
							coefficient, field_power);
				buf_read[k] = buf_read[k] ^ encoded_data;
                                encoded_data = (char)gf_mul((unsigned char)buf[k],
                                                        coefficient, field_power);
                                buf_read[k] = buf_read[k] ^ encoded_data;
			}

                        //write coded block
                        // Cache Start
                        retstat = cacheLayer->DiskWrite(code_disk_id,buf_read,size,code_block_no*block_size);
                        // Cache End
                        //printf("***msr: retstat=%d\n",retstat);
		}
                }

                free(buf_read);
                free(buf2);
        }

        block_written.disk_id = disk_id;
        block_written.block_no = block_no;

        return block_written;
}


/*************************************************************************
 * Decoding functions
 *************************************************************************/


/*
 * decoding_default: no decoding
 *
 * @param disk_id: disk id
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer
 * @param offset: offset of data on disk
 *
 *@return: size of successful read (in bytes)
 */
int CodingLayer::decoding_default(int disk_id, char* buf, long long size, long long offset)
{
	if(NCFS_DATA->disk_status[disk_id] == 0)
		return cacheLayer->DiskRead(disk_id,buf,size,offset);
	else {
		printf("Raid %d: Disk %d failed\n",NCFS_DATA->disk_raid_type,disk_id);
		return -1;
	}
	AbnormalError();
	return -1;
}


/*
 * Decoding_JBOD: JBOD decode
 *
 * @param disk_id: disk id
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer
 * @param offset: offset of data on disk
 *
 * @return: size of successful decoding (in bytes)
 */
int CodingLayer::decoding_jbod(int disk_id, char* buf, long long size, long long offset)
{
	return decoding_default(disk_id,buf,size,offset);
}


/*
 * decoding_raid0: RAID 0 decode
 *
 * @param disk_id: disk id
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer
 * @param offset: offset of data on disk
 *
 * @return: size of successful decoding (in bytes)
 */
int CodingLayer::decoding_raid0(int disk_id, char* buf, long long size, long long offset){
	return decoding_default(disk_id,buf,size,offset);
}


/*
 * decoding_raid1: RAID 1 decode
 *
 * @param disk_id: disk id
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer
 * @param offset: offset of data on disk
 *
 * @return: size of successful decoding (in bytes)
 */
int CodingLayer::decoding_raid1(int disk_id, char* buf, long long size, long long offset)
{
	if(NCFS_DATA->disk_status[disk_id] == 0)
		return cacheLayer->DiskRead(disk_id,buf,size,offset);
	else {
		int mirror_disk_id = NCFS_DATA->disk_total_num - disk_id - 1;
		if(NCFS_DATA->disk_status[mirror_disk_id] == 0){
			return cacheLayer->DiskRead(mirror_disk_id,buf,size,offset);
		} else {
			printf("Raid 1 both disk %d and mirror disk %d\n",disk_id,mirror_disk_id);
			return -1;
		}
	}
	AbnormalError();
	return -1;
} 


/*
 * decoding_raid4: RAID 4 decode
 *
 * @param disk_id: disk id               
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer
 * @param offset: offset of data on disk
 *
 * @return: size of successful decoding (in bytes)
 */
int CodingLayer::decoding_raid4(int disk_id, char* buf, long long size, long long offset)
{        
	if(NCFS_DATA->disk_status[disk_id] == 0)
		return cacheLayer->DiskRead(disk_id,buf,size,offset);
	else {
		int retstat;
		char* temp_buf;
		int i;
		long long j;
		temp_buf = (char*)malloc(sizeof(char)*size);
		memset(temp_buf, 0, size);
		memset(buf, 0, size);
		int* inttemp_buf = (int*)temp_buf;
		int* intbuf = (int*)buf;
		for(i = 0; i < NCFS_DATA->disk_total_num; ++i){
			if(i != disk_id){
				if(NCFS_DATA->disk_status[i] != 0){
					printf("Raid 4 both disk %d and %d are failed\n",disk_id,i);
					return -1;
				}
				retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);
				for(j = 0; j < (long long)(size * sizeof(char)/sizeof(int)); ++j)
					intbuf[j] = intbuf[j] ^ inttemp_buf[j];
			}
		}
		free(temp_buf);
		return size;
	}
	AbnormalError();             
	return -1;
}


/*
 * decoding_raid5: RAID 5 decode
 *
 * @param disk_id: disk id               
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer
 * @param offset: offset of data on disk
 *
 * @return: size of successful decoding (in bytes)
 */
int CodingLayer::decoding_raid5(int disk_id, char* buf, long long size, long long offset)
{
	int retstat;
        struct timeval t1, t2, t3;
        double duration;

	if(NCFS_DATA->disk_status[disk_id] == 0){
                if (NCFS_DATA->run_experiment == 1){
                	gettimeofday(&t1,NULL);
                }

		retstat = cacheLayer->DiskRead(disk_id,buf,size,offset);

                if (NCFS_DATA->run_experiment == 1){
                	gettimeofday(&t2,NULL);

                	//duration = (t2.tv_sec - t1.tv_sec) + 
                        //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                        //NCFS_DATA->diskread_time += duration;
                }

		return retstat;
	}
	else {
		//printf("Degraded Read\n");
		char* temp_buf;
		int i;
		long long j;
		temp_buf = (char*)malloc(sizeof(char)*size);
		memset(temp_buf, 0, size);
		memset(buf, 0, size);
		int* inttemp_buf = (int*)temp_buf;
		int* intbuf = (int*)buf;
		for(i = 0; i < NCFS_DATA->disk_total_num; ++i){
			if(i != disk_id){
				if(NCFS_DATA->disk_status[i] != 0){
					printf("Raid 5 both disk %d and %d are failed\n",disk_id,i);
					free(temp_buf);
					return -1;
				}

				if (NCFS_DATA->run_experiment == 1){
                                	gettimeofday(&t1,NULL);
                        	}

				retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);

				if (NCFS_DATA->run_experiment == 1){
                                	gettimeofday(&t2,NULL);
                        	}

				for(j = 0; j < (long long)(size * sizeof(char) / sizeof(int)); ++j)
					intbuf[j] = intbuf[j] ^ inttemp_buf[j];

				if (NCFS_DATA->run_experiment == 1){
                                	gettimeofday(&t3,NULL);

                                	//duration = (t2.tv_sec - t1.tv_sec) + 
                                       	//	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                	//NCFS_DATA->diskread_time += duration;

                                	duration = (t3.tv_sec - t2.tv_sec) + 
                                        	(t3.tv_usec-t2.tv_usec)/1000000.0;
                                	NCFS_DATA->decoding_time += duration;
                        	}
			}
		}
		free(temp_buf);
		return size;
	}
	AbnormalError();
	return -1;
}


/*
 * decoding_raid6: RAID 6 decode
 *
 * @param disk_id: disk id
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer
 * @param offset: offset of data on disk
 *
 * @return: size of successful decoding (in bytes)
 */
int CodingLayer::decoding_raid6(int disk_id, char* buf, long long size, long long offset)
{

        int retstat;
        char* temp_buf;
        int i;
        long long j;
        int disk_failed_no;
        //int disk_second_id;     //id of second failed disk
        int disk_another_failed_id; //id of second failed disk

        int code_disk_id, parity_disk_id;
        int block_size;
        long long block_no;
        int disk_total_num;
        int data_disk_coeff;
        char temp_char;

        int g1, g2, g12;
        char* P_temp;
	char* Q_temp;

        struct timeval t1, t2, t3;
        double duration;

        temp_buf = (char*)malloc(sizeof(char)*size);        
        memset(temp_buf, 0, size);

        P_temp = (char*)malloc(sizeof(char)*size);        
        memset(P_temp, 0, size);        

        Q_temp = (char*)malloc(sizeof(char)*size);        
        memset(Q_temp, 0, size);        
        
        memset(buf, 0, size);        
        int* inttemp_buf = (int*)temp_buf;                      
        int* intbuf = (int*)buf;      

        disk_total_num = NCFS_DATA->disk_total_num;                 
        block_size = NCFS_DATA->disk_block_size;
        block_no = offset / block_size;

        code_disk_id = disk_total_num -1 - (block_no % disk_total_num);
        parity_disk_id = disk_total_num - 1 - ((block_no+1) % disk_total_num);
        //code_disk_id = disk_total_num - 1;
        //parity_disk_id = disk_total_num - 2;


        if(NCFS_DATA->disk_status[disk_id] == 0){ 
                if (NCFS_DATA->run_experiment == 1){
                	gettimeofday(&t1,NULL);
                }

                retstat = cacheLayer->DiskRead(disk_id,buf,size,offset);

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t2,NULL);

                        //duration = (t2.tv_sec - t1.tv_sec) + 
                        //        (t2.tv_usec-t1.tv_usec)/1000000.0;
                        //NCFS_DATA->diskread_time += duration;
                }

		free(temp_buf);
		free(P_temp);
		free(Q_temp);

		return retstat;
	}
        else {
                //check the number of failed disks
                disk_failed_no = 0;
                for (i=0; i < disk_total_num; i++){
                        if (NCFS_DATA->disk_status[i] == 1){
                                (disk_failed_no)++;
                                if (i != disk_id){
                                        //disk_second_id = i;
					disk_another_failed_id = i;
                                }
                        }
                }


                if ( ((disk_failed_no == 1) && (disk_id != code_disk_id)) ||
                        ((disk_failed_no == 2) && (disk_another_failed_id == code_disk_id)) ){
                //case of single non-Q disk fail (D or P), or two-disk (data disk + Q disk) fail (D + Q)
                        for(i = 0; i < disk_total_num; ++i){
                                if ((i != disk_id) && (i != code_disk_id)){
                                        if(NCFS_DATA->disk_status[i] != 0){
                                                printf("Raid 6 both disk %d and %d are failed\n",disk_id,i);
                                                return -1;
                                        }

					if (NCFS_DATA->run_experiment == 1){
                                        	gettimeofday(&t1,NULL);
                                	}

                                        retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);

                                        if (NCFS_DATA->run_experiment == 1){
                                                gettimeofday(&t2,NULL);
                                        }

                                        for(j = 0; j < (long long)(size * sizeof(char) / sizeof(int)); ++j){
                                                intbuf[j] = intbuf[j] ^ inttemp_buf[j];
                                        }

                                	if (NCFS_DATA->run_experiment == 1){
                                        	gettimeofday(&t3,NULL);

                                        	//duration = (t2.tv_sec - t1.tv_sec) + 
                                                //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                        	//NCFS_DATA->diskread_time += duration;

                                        	duration = (t3.tv_sec - t2.tv_sec) + 
                                                	(t3.tv_usec-t2.tv_usec)/1000000.0;
                                        	NCFS_DATA->decoding_time += duration;
                                	}
                                }
                        }
                }
		else if ( ((disk_failed_no == 1) && (disk_id == code_disk_id))
				|| ((disk_failed_no == 2) && (disk_id == code_disk_id) 
					&& (disk_another_failed_id == parity_disk_id)) ){
		//case of "single Q disk fails" (Q) or "Q + P disk failure" (Q + P)

			//Calculate Q
                	for (i=0; i < disk_total_num; i++){
                        	if ((i != parity_disk_id) && (i != code_disk_id)){

                                        if (NCFS_DATA->run_experiment == 1){
                                                gettimeofday(&t1,NULL);
                                        }

                                	//Cache Start
                                	retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);
                                	//Cache End

                                        if (NCFS_DATA->run_experiment == 1){
                                                gettimeofday(&t2,NULL);
                                        }

                                	for (j=0; j < size; j++){
                                        	//calculate the coefficient of the data block
                                        	data_disk_coeff = i;

                                        	if (i > code_disk_id){
                                                	(data_disk_coeff)--;
                                        	}
                                        	if (i > parity_disk_id){
                                                	(data_disk_coeff)--;
                                        	}
                                        	data_disk_coeff = disk_total_num - 3 - data_disk_coeff;
                                       		data_disk_coeff = gf_get_coefficient(data_disk_coeff, field_power);

                                        	//calculate code block Q
                                        	buf[j] = buf[j] ^
                                        	(char)gf_mul((unsigned char)temp_buf[j],data_disk_coeff,field_power);
                                	}

                                        if (NCFS_DATA->run_experiment == 1){
                                                gettimeofday(&t3,NULL);

                                                //duration = (t2.tv_sec - t1.tv_sec) + 
                                                //        (t2.tv_usec-t1.tv_usec)/1000000.0;
                                                //NCFS_DATA->diskread_time += duration;

                                                duration = (t3.tv_sec - t2.tv_sec) + 
                                                        (t3.tv_usec-t2.tv_usec)/1000000.0;
                                                NCFS_DATA->decoding_time += duration;
                                        }
                        	}
                	}
		}
                else if (disk_failed_no == 2){
                //case of two-disk fail (data disk + non Q disk) (disk_id and disk_second_id)
                //compute Q' to restore data

			if ((disk_id == parity_disk_id) && (disk_another_failed_id != code_disk_id)){
			//case of P disk + data disk fail (P + D)

				//calculate Q'
                                for (i=0; i < disk_total_num; i++){
                                        if ((i != disk_id) && (i != disk_another_failed_id) &&
                                        (i != parity_disk_id) && ( i != code_disk_id) ){

                                        	if (NCFS_DATA->run_experiment == 1){
                                                	gettimeofday(&t1,NULL);
                                        	}

                                                retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t2,NULL);
                                                }

                                                //calculate the coefficient of the data block
                                                data_disk_coeff = i;

                                                if (i > code_disk_id){
                                                        (data_disk_coeff)--;
                                                }
                                                if (i > parity_disk_id){
                                                        (data_disk_coeff)--;
                                                }
                                                data_disk_coeff = disk_total_num - 3 - data_disk_coeff;
                                                data_disk_coeff = gf_get_coefficient(data_disk_coeff, field_power);

                                                for (j=0; j < size; j++){
                                                        buf[j] = buf[j] ^
                                                        (char)gf_mul((unsigned char)temp_buf[j],data_disk_coeff,field_power);
                                                }

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t3,NULL);

                                                	//duration = (t2.tv_sec - t1.tv_sec) + 
                                                        //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                                	//NCFS_DATA->diskread_time += duration;

                                                	duration = (t3.tv_sec - t2.tv_sec) + 
                                                        	(t3.tv_usec-t2.tv_usec)/1000000.0;
                                                	NCFS_DATA->decoding_time += duration;
                                                }
                                        }
                                }

                                //calculate Q xor Q'

                                if (NCFS_DATA->run_experiment == 1){
                                	gettimeofday(&t1,NULL);
                                }

                                retstat = cacheLayer->DiskRead(code_disk_id,temp_buf,size,offset);

                                if (NCFS_DATA->run_experiment == 1){
                                        gettimeofday(&t2,NULL);
                                }

                                for (j=0; j < size; j++){
                                        buf[j] = buf[j] ^ temp_buf[j];
                                }

                                //calculate the coefficient of the data block
                                data_disk_coeff = disk_id;

                                if (disk_id > code_disk_id){
                                        (data_disk_coeff)--;
                                }
                                if (disk_id > parity_disk_id){
                                        (data_disk_coeff)--;
                                }
                                data_disk_coeff = disk_total_num - 3 - data_disk_coeff;
                                data_disk_coeff = gf_get_coefficient(data_disk_coeff, field_power);


                                //decode the origianl data block
                                for (j=0; j < size; j++){
                                        temp_char = buf[j];
                                        buf[j] = (char)gf_div((unsigned char)temp_char,data_disk_coeff,field_power);
                                }

				if (NCFS_DATA->run_experiment == 1){
                                        gettimeofday(&t3,NULL);

                                        //duration = (t2.tv_sec - t1.tv_sec) + 
                                        //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                        //NCFS_DATA->diskread_time += duration;

                                        duration = (t3.tv_sec - t2.tv_sec) + 
                                        	(t3.tv_usec-t2.tv_usec)/1000000.0;
                                        NCFS_DATA->decoding_time += duration;
                                }

				//find P
                        	for(i = 0; i < disk_total_num; ++i){
                                	if ((i != disk_id) && (i != code_disk_id) && (i != disk_another_failed_id)){
                                        	if(NCFS_DATA->disk_status[i] != 0){
                                                	printf("Raid 6 both disk %d and %d are failed\n",disk_id,i);
                                                	return -1;
                                        	}

                                		if (NCFS_DATA->run_experiment == 1){
                                        		gettimeofday(&t1,NULL);
                                		}

                                        	retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t2,NULL);
                                                }

                                        	for(j = 0; j < (long long)(size * sizeof(char) / sizeof(int)); ++j){
                                                	intbuf[j] = intbuf[j] ^ inttemp_buf[j];
                                        	}

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t3,NULL);

                                                        //duration = (t2.tv_sec - t1.tv_sec) + 
                                                        //        (t2.tv_usec-t1.tv_usec)/1000000.0;
                                                        //NCFS_DATA->diskread_time += duration;

                                                        duration = (t3.tv_sec - t2.tv_sec) + 
                                                                (t3.tv_usec-t2.tv_usec)/1000000.0;
                                                        NCFS_DATA->decoding_time += duration;
                                                }
                                	}
                        	}
			}
                        else if ((disk_id == code_disk_id) && (disk_another_failed_id != parity_disk_id)){
                        //case of Q disk + data disk fail (Q + D)

                                //find D from P
                                for(i = 0; i < disk_total_num; ++i){
                                        if ((i != disk_id) && (i != disk_another_failed_id)){
                                                if(NCFS_DATA->disk_status[i] != 0){
                                                        printf("Raid 6 both disk %d and %d are failed\n",disk_id,i);
                                                        return -1;
                                                }

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t1,NULL);
                                                }

                                                retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t2,NULL);
                                                }

                                                for(j = 0; j < (long long)(size * sizeof(char) / sizeof(int)); ++j){
                                                        intbuf[j] = intbuf[j] ^ inttemp_buf[j];
                                                }

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t3,NULL);

                                                        //duration = (t2.tv_sec - t1.tv_sec) +
                                                        //        (t2.tv_usec-t1.tv_usec)/1000000.0;
                                                        //NCFS_DATA->diskread_time += duration;

                                                        duration = (t3.tv_sec - t2.tv_sec) +
                                                                (t3.tv_usec-t2.tv_usec)/1000000.0;
                                                        NCFS_DATA->decoding_time += duration;
                                                }
                                        }
                                }

                        	//calculate Q
                        	for (i=0; i < disk_total_num; i++){
                                	if ((i != parity_disk_id) && (i != code_disk_id)){
						if (i != disk_another_failed_id){
                                                	if (NCFS_DATA->run_experiment == 1){
                                                        	gettimeofday(&t1,NULL);
                                                	}

                                        		//Cache Start
                                        		retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);
                                        		//Cache End

                                                	if (NCFS_DATA->run_experiment == 1){
                                                        	gettimeofday(&t2,NULL);

                                                        	//duration = (t2.tv_sec - t1.tv_sec) + 
                                                                //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                                        	//NCFS_DATA->diskread_time += duration;
                                                	}
						}
						else{ //use the recovered D
                                                	for(j = 0; j < size; j++){
                                                        	temp_buf[j] = buf[j];
                                                	}
						}

                                                if (NCFS_DATA->run_experiment == 1){
                                                	gettimeofday(&t1,NULL);
                                                }

                                        	for (j=0; j < size; j++){
                                                	//calculate the coefficient of the data block
                                                	data_disk_coeff = i;

                                                	if (i > code_disk_id){
                                                        	(data_disk_coeff)--;
                                                	}
                                                	if (i > parity_disk_id){
                                                        	(data_disk_coeff)--;
                                                	}
                                                	data_disk_coeff = disk_total_num - 3 - data_disk_coeff;
                                                	data_disk_coeff = gf_get_coefficient(data_disk_coeff, field_power);

                                                	//calculate code block Q
                                                	Q_temp[j] = Q_temp[j] ^
                                                	(char)gf_mul((unsigned char)temp_buf[j],data_disk_coeff,field_power);
                                        	}

                                                if (NCFS_DATA->run_experiment == 1){
                                                	gettimeofday(&t2,NULL);

                                                        //duration = (t2.tv_sec - t1.tv_sec) +
                                                        //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                                        //NCFS_DATA->diskread_time += duration;
                                                }
                                	}
                        	}

				for (j=0; j<size; j++){
					buf[j] = Q_temp[j];
				}
                        }
                        else if (disk_another_failed_id == parity_disk_id){
                        //case of data disk + P disk fail (D + P)

                                //calculate Q'
                                for (i=0; i < disk_total_num; i++){
                                        if ((i != disk_id) && (i != disk_another_failed_id) &&
                                        (i != parity_disk_id) && ( i != code_disk_id) ){

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t1,NULL);
                                                }

                                                retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t2,NULL);
                                                }

                                                //calculate the coefficient of the data block
                                                data_disk_coeff = i;

                                                if (i > code_disk_id){
                                                        (data_disk_coeff)--;
                                                }
                                                if (i > parity_disk_id){
                                                        (data_disk_coeff)--;
                                                }
                                                data_disk_coeff = disk_total_num - 3 - data_disk_coeff;
                                                data_disk_coeff = gf_get_coefficient(data_disk_coeff, field_power);

                                                for (j=0; j < size; j++){
                                                        temp_char = buf[j];
							buf[j] = buf[j] ^
                                                       	(char)gf_mul((unsigned char)temp_buf[j],data_disk_coeff,field_power);
                                                }

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t3,NULL);

                                                        //duration = (t2.tv_sec - t1.tv_sec) +
                                                        //        (t2.tv_usec-t1.tv_usec)/1000000.0;
                                                        //NCFS_DATA->diskread_time += duration;

                                                        duration = (t3.tv_sec - t2.tv_sec) +
                                                                (t3.tv_usec-t2.tv_usec)/1000000.0;
                                                        NCFS_DATA->decoding_time += duration;
                                                }
                                        }
                                }

                                //calculate Q xor Q'

                                if (NCFS_DATA->run_experiment == 1){
                                	gettimeofday(&t1,NULL);
                                }

                                retstat = cacheLayer->DiskRead(code_disk_id,temp_buf,size,offset);

                                if (NCFS_DATA->run_experiment == 1){
                                        gettimeofday(&t2,NULL);
                                }

                                for (j=0; j < size; j++){
                                        buf[j] = buf[j] ^ temp_buf[j];
                                }

                                //calculate the coefficient of the data block
                                data_disk_coeff = disk_id;

                                if (disk_id > code_disk_id){
                                        (data_disk_coeff)--;
                                }
                                if (disk_id > parity_disk_id){
                                        (data_disk_coeff)--;
                                }
                                data_disk_coeff = disk_total_num - 3 - data_disk_coeff;
                                data_disk_coeff = gf_get_coefficient(data_disk_coeff, field_power);


                                //decode the origianl data block
                                for (j=0; j < size; j++){
                                        temp_char = buf[j];
                                        buf[j] = (char)gf_div((unsigned char)temp_char,data_disk_coeff,field_power);
                                }

                                if (NCFS_DATA->run_experiment == 1){
                                        gettimeofday(&t3,NULL);

                                        //duration = (t2.tv_sec - t1.tv_sec) +
                                        //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                        //NCFS_DATA->diskread_time += duration;

                                        duration = (t3.tv_sec - t2.tv_sec) +
                                        	(t3.tv_usec-t2.tv_usec)/1000000.0;
                                	NCFS_DATA->decoding_time += duration;
                                }

                                //return retstat;
                        }
                        else{
                        // two data disk fail (D + D)

                                if (NCFS_DATA->run_experiment == 1){
                                        gettimeofday(&t1,NULL);
                                }

                                //calculate g1
                                g1 = disk_id;
                                if (g1 > code_disk_id){
                                        (g1)--;
                                }
                                if (g1 > parity_disk_id){
                                        (g1)--;
                                }
                                g1 = disk_total_num - 3 - g1;
                                g1 = gf_get_coefficient(g1, field_power);
                                //printf("g1=%d \n",g1);

                                //calculate g2
                                g2 = disk_another_failed_id;
                                if (g2 > code_disk_id){
                                        (g2)--;
                                }
                                if (g2 > parity_disk_id){
                                        (g2)--;
                                }
                                g2 = disk_total_num - 3 - g2;
                                g2 = gf_get_coefficient(g2, field_power);
                                //printf("g2=%d \n",g2);

                                //calculate g12
                                g12 = g1 ^ g2;

                                //calculate P'
                                //for (j=0; j<BLOCKSIZE; j++){
                                //        P_temp[j] = 0;
                                //}

                                if (NCFS_DATA->run_experiment == 1){
                                        gettimeofday(&t2,NULL);

                                        duration = (t2.tv_sec - t1.tv_sec) +
                                                (t2.tv_usec-t1.tv_usec)/1000000.0;
                                        NCFS_DATA->decoding_time += duration;
                                }

                                for (i=0; i<disk_total_num; i++){
                                        if ( (i != disk_id) && (i != disk_another_failed_id) &&
                                                (i != parity_disk_id) && (i != code_disk_id) )
                                        {
                               			if (NCFS_DATA->run_experiment == 1){
                                        		gettimeofday(&t1,NULL);
                                		}

						retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t2,NULL);
                                                }

                                                for (j=0; j<size; j++){
                                                        P_temp[j] = P_temp[j] ^ temp_buf[j];
                                                }

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t3,NULL);

                                        		//duration = (t2.tv_sec - t1.tv_sec) +
                                                	//	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                        		//NCFS_DATA->diskread_time += duration;

                                        		duration = (t3.tv_sec - t2.tv_sec) +
                                                		(t3.tv_usec-t2.tv_usec)/1000000.0;
                                        		NCFS_DATA->decoding_time += duration;
                                                }
                                        }
                                }

                                if (NCFS_DATA->run_experiment == 1){
                                	gettimeofday(&t1,NULL);
                                }

				retstat = cacheLayer->DiskRead(parity_disk_id,temp_buf,size,offset);

                                if (NCFS_DATA->run_experiment == 1){
                                	gettimeofday(&t2,NULL);
                                }

                                for (j=0; j<size; j++){
                                        P_temp[j] = P_temp[j] ^ temp_buf[j];
                                        //P_temp = P' xor P
                                }

                                if (NCFS_DATA->run_experiment == 1){
                                	gettimeofday(&t3,NULL);

                                        //duration = (t2.tv_sec - t1.tv_sec) +
                                        //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                        //NCFS_DATA->diskread_time += duration;

                                        duration = (t3.tv_sec - t2.tv_sec) +
                                        	(t3.tv_usec-t2.tv_usec)/1000000.0;
                                        NCFS_DATA->decoding_time += duration;
                                }

                                //calculate Q'
                                //for (j=0; j<BLOCKSIZE; j++){
                                //        Q_temp[j] = 0;

                                for (i=0; i<disk_total_num; i++){
                                        if ( ((i != disk_id) && (i != disk_another_failed_id))
                                        && (i != parity_disk_id) && (i != code_disk_id) ){

                                                //calculate the coefficient of the data block
                                                data_disk_coeff = i;

                                                if (i > code_disk_id){
                                                        (data_disk_coeff)--;
                                                }
                                                if (i > parity_disk_id){
                                                        (data_disk_coeff)--;
                                                }
                                                data_disk_coeff = disk_total_num - 3 - data_disk_coeff;
                                                data_disk_coeff = gf_get_coefficient(data_disk_coeff, field_power);

                                                //printf("\ndata disk coefficient = %d\n",data_disk_coeff);       

                                		if (NCFS_DATA->run_experiment == 1){
                                        		gettimeofday(&t1,NULL);
                                		}

						retstat = cacheLayer->DiskRead(i,temp_buf,size,offset);

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t2,NULL);
                                                }

                                                for (j=0; j < size; j++){
                                                        temp_char = Q_temp[j];
                                                        Q_temp[j] = temp_char ^
                                                        (char)gf_mul((unsigned char)temp_buf[j],data_disk_coeff,field_power);
                                                }

                                                if (NCFS_DATA->run_experiment == 1){
                                                        gettimeofday(&t3,NULL);

                                        		//duration = (t2.tv_sec - t1.tv_sec) +
                                                	//	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                        		//NCFS_DATA->diskread_time += duration;

                                        		duration = (t3.tv_sec - t2.tv_sec) +
                                                		(t3.tv_usec-t2.tv_usec)/1000000.0;
                                        		NCFS_DATA->decoding_time += duration;
                                                }
                                        }

                                        //for (j=0; j<size; j++){
                                        //        printf("Q'=%c(%d) ",Q_temp[j],Q_temp[j]);
                                        //}

                                	//calculate D

                                	if (NCFS_DATA->run_experiment == 1){
                                        	gettimeofday(&t1,NULL);
                                	}

					retstat = cacheLayer->DiskRead(code_disk_id,temp_buf,size,offset);

                                        if (NCFS_DATA->run_experiment == 1){
                                                gettimeofday(&t2,NULL);
                                        }

                                	for (j=0; j<size; j++){
                                        	temp_char = (char)(gf_mul(g2,(unsigned char)P_temp[j],field_power) 
                                                                ^ Q_temp[j] ^ temp_buf[j]);
                                        	buf[j] = (char)gf_div((unsigned char)temp_char, g12, field_power);
                                	}

                                        if (NCFS_DATA->run_experiment == 1){
                                                gettimeofday(&t3,NULL);

                                                //duration = (t2.tv_sec - t1.tv_sec) +
                                                //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                                                //NCFS_DATA->diskread_time += duration;

                                                duration = (t3.tv_sec - t2.tv_sec) +
                                                	(t3.tv_usec-t2.tv_usec)/1000000.0;
                                                NCFS_DATA->decoding_time += duration;
                                        }
                                }
                        }
                }
                else{
                        printf("Raid 6 number of failed disks larger than 2.\n");
                        return -1;
                }

                free(temp_buf);
		free(P_temp);
		free(Q_temp);
                return size;
        }
        AbnormalError();

        return -1;
}



/*
 * decoding_mbr: MBR decode
 *
 * @param disk_id: disk id
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer
 * @param offset: offset of data on disk
 *
 * @return: size of successful decoding (in bytes)
 */
int CodingLayer::decoding_mbr(int disk_id, char* buf, long long size, long long offset)
{
	int retstat;
        struct timeval t1, t2, t3, t4;
        double duration;

        if(NCFS_DATA->disk_status[disk_id] == 0){
                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t1,NULL);
                }

                retstat = cacheLayer->DiskRead(disk_id,buf,size,offset);

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t2,NULL);

                        //duration = (t2.tv_sec - t1.tv_sec) + 
                        //        (t2.tv_usec-t1.tv_usec)/1000000.0;
                        //NCFS_DATA->diskread_time += duration;
                }

		return retstat;
	}
        else {
                //printf("Degraded Read\n");
                char* temp_buf;
                int i;
                long long j;
                temp_buf = (char*)malloc(sizeof(char)*size);
                memset(temp_buf, 0, size);
                memset(buf, 0, size);
                int* inttemp_buf = (int*)temp_buf;
                int* intbuf = (int*)buf;

		int mbr_segment_size;
		int mbr_segment_id;
		int mbr_block_id;
		long long block_no;
		int dup_disk_id, dup_block_no;
		int temp_mbr_block_id, temp_disk_id, temp_block_no;
		int flag_dup;
		int disk_failed_no, disk_another_failed_id;
		long long offset_read;
		int temp;

                //check the number of failed disks
                disk_failed_no = 0;
		disk_another_failed_id = -1;
                for (i=0; i < NCFS_DATA->disk_total_num; i++){
                        if (NCFS_DATA->disk_status[i] == 1){
                                (disk_failed_no)++;
                                if (i != disk_id){
                                        disk_another_failed_id = i;
                                }
                        }
                }

		mbr_segment_size = NCFS_DATA->mbr_segment_size;
		block_no = (long long)(offset / NCFS_DATA->disk_block_size);
		mbr_segment_id = (int)(block_no / mbr_segment_size);

                if (NCFS_DATA->run_experiment == 1){
                	gettimeofday(&t1,NULL);
                }

		flag_dup = 0;
		mbr_block_id = mbr_find_block_id(disk_id, block_no, mbr_segment_size);
		if (mbr_block_id == -1){
			flag_dup = 1;
			mbr_block_id = mbr_find_dup_block_id(disk_id, block_no, mbr_segment_size);
		}

		if (flag_dup == 0){
			dup_disk_id = mbr_get_dup_disk_id(mbr_block_id,mbr_segment_size);
			dup_block_no = mbr_get_dup_block_no(disk_id,mbr_block_id,mbr_segment_size);
			dup_block_no = dup_block_no + mbr_segment_id * mbr_segment_size;
		}
		else{ //if flag_dup = 1
                        dup_disk_id = mbr_get_disk_id(mbr_block_id,mbr_segment_size);
                        dup_block_no = mbr_get_block_no(dup_disk_id,mbr_block_id,mbr_segment_size);
                        dup_block_no = dup_block_no + mbr_segment_id * mbr_segment_size;
		}

                if (NCFS_DATA->run_experiment == 1){
                        gettimeofday(&t2,NULL);

                        duration = (t2.tv_sec - t1.tv_sec) +
                        	(t2.tv_usec-t1.tv_usec)/1000000.0;
                        NCFS_DATA->decoding_time += duration;
                }


		if (dup_disk_id == disk_another_failed_id){
		//Use coded block to decode
		//printf("***MBR: decode by coded block. mbr_block_id=%d\n",mbr_block_id);
			for (i=0; i<((NCFS_DATA->mbr_m) + (NCFS_DATA->mbr_c)); i++){
			//currently use XOR, and supports only single coded block
				if (i != mbr_block_id){
					if (NCFS_DATA->run_experiment == 1){
                                		gettimeofday(&t1,NULL);
                        		}

					temp_mbr_block_id = i;
					temp_disk_id = mbr_get_disk_id(temp_mbr_block_id, mbr_segment_size);
					temp_block_no = mbr_get_block_no(temp_disk_id, temp_mbr_block_id,
								mbr_segment_size);

					if (NCFS_DATA->disk_status[temp_disk_id] == 1){
					//original data block fails, use duplicated block
						temp = temp_disk_id;
						temp_disk_id = mbr_get_dup_disk_id(temp_mbr_block_id, 
										mbr_segment_size);
						temp_block_no = mbr_get_dup_block_no(temp, temp_mbr_block_id,
										mbr_segment_size);
					}

					temp_block_no = temp_block_no + mbr_segment_id * mbr_segment_size;
					offset_read = (long long)(temp_block_no * (NCFS_DATA->disk_block_size));

                                        if (NCFS_DATA->run_experiment == 1){
                                                gettimeofday(&t2,NULL);
                                        }

					retstat = cacheLayer->DiskRead(temp_disk_id,temp_buf,size,offset_read);

                                        if (NCFS_DATA->run_experiment == 1){
                                                gettimeofday(&t3,NULL);
                                        }


					if (retstat < 0 ){
						return -1;
					}

                                	for(j = 0; j < (long long)(size * sizeof(char) / sizeof(int)); ++j){
                                                intbuf[j] = intbuf[j] ^ inttemp_buf[j];
					}

                                        if (NCFS_DATA->run_experiment == 1){
                                                gettimeofday(&t4,NULL);

                                		duration = (t2.tv_sec - t1.tv_sec) +
                                        		(t2.tv_usec-t1.tv_usec)/1000000.0;
                                		NCFS_DATA->decoding_time += duration;

                                                //duration = (t3.tv_sec - t2.tv_sec) +
                                                //        (t3.tv_usec-t2.tv_usec)/1000000.0;
                                                //NCFS_DATA->diskread_time += duration;

                                                duration = (t4.tv_sec - t3.tv_sec) +
                                                        (t4.tv_usec-t3.tv_usec)/1000000.0;
                                                NCFS_DATA->decoding_time += duration;
                                        }
				}
			}
		}
		else{
		//Use duplicated block to decode
			offset_read = (long long)(dup_block_no * (NCFS_DATA->disk_block_size));

                	if (NCFS_DATA->run_experiment == 1){
                        	gettimeofday(&t1,NULL);
                	}

			retstat = cacheLayer->DiskRead(dup_disk_id,buf,size,offset_read);

                	if (NCFS_DATA->run_experiment == 1){
                        	gettimeofday(&t2,NULL);

                        	//duration = (t2.tv_sec - t1.tv_sec) +
                                //	(t2.tv_usec-t1.tv_usec)/1000000.0;
                        	//NCFS_DATA->diskread_time += duration;
                	}

			return retstat;
		}

                free(temp_buf);
                return size;
        }


        AbnormalError();
        return -1;
}



/*
 * decoding_msr: MSR decode
 *
 * @param disk_id: disk id               
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer                  
 * @param offset: offset of data on disk
 *
 * @return: size of successful decoding (in bytes)
 */
int CodingLayer::decoding_msr(int disk_id, char* buf, long long size, long long offset)
{
        if(NCFS_DATA->disk_status[disk_id] == 0)
                return cacheLayer->DiskRead(disk_id,buf,size,offset);
	else{
                //printf("Degraded Read\n");
                int retstat;
                char* temp_buf;
		char* temp_buf2;
                int i, j, k;

		int msr_n, msr_k, msr_m, msr_c;
                int msr_segment_size;
                int msr_segment_id;
                int msr_block_id;
                long long block_no;
                int code_block_id, code_disk_id, code_block_no;
                int temp_disk_id, temp_block_no;
                int flag_code;
                int disk_failed_no, disk_another_failed_id;
                long long offset_read;
		int coefficient;

                temp_buf = (char*)malloc(sizeof(char)*size);
                temp_buf2 = (char*)malloc(sizeof(char)*size);
                memset(temp_buf, 0, size);
                memset(temp_buf2, 0, size);
                memset(buf, 0, size);


                //check the number of failed disks
                disk_failed_no = 0; 
                disk_another_failed_id = -1;             
                for (i=0; i < NCFS_DATA->disk_total_num; i++){                       
                        if (NCFS_DATA->disk_status[i] == 1){
                                (disk_failed_no)++;            
                                if (i != disk_id){
                                        disk_another_failed_id = i;                
                                }
                        }                        
                } 

		msr_n = NCFS_DATA->msr_n;
		msr_k = NCFS_DATA->msr_k;
		msr_m = NCFS_DATA->msr_m;
		msr_c = NCFS_DATA->msr_c;
                msr_segment_size = NCFS_DATA->msr_segment_size;                  
                block_no = (long long)(offset / NCFS_DATA->disk_block_size);
                msr_segment_id = (int)(block_no / msr_segment_size);

                flag_code = 0;
                msr_block_id = msr_find_block_id(disk_id, block_no, msr_segment_size, msr_n, msr_k);
                if (msr_block_id == -1){
                        flag_code = 1;
                        code_block_id = msr_find_code_block_id(disk_id, block_no, msr_segment_size, msr_n, msr_k);
                }
		else{
			code_block_id = msr_find_single_decode_block_id(msr_block_id, msr_segment_size);
		}

                code_disk_id = msr_get_code_disk_id(code_block_id, msr_segment_size, msr_n, msr_k);
                code_block_no = msr_get_code_block_no(msr_block_id, msr_segment_size);
                code_block_no = code_block_no + msr_segment_id * msr_segment_size;

		if ((disk_failed_no == 1) && (flag_code == 0)){ //1 data node failure. recover data disk.
			offset_read = code_block_no * (NCFS_DATA->disk_block_size);
			retstat = cacheLayer->DiskRead(code_disk_id,temp_buf,size,offset_read);

			for (j=0; j<msr_m; j++){
				coefficient =msr_coding_matrix[code_block_id][j];
				if ((j != msr_block_id) && (coefficient != 0)){
					temp_disk_id = msr_get_disk_id(j, msr_segment_size, msr_n, msr_k);
					temp_block_no = msr_get_block_no(j, msr_segment_size);
					temp_block_no = temp_block_no + msr_segment_id * msr_segment_size;
                                	offset_read = temp_block_no * (NCFS_DATA->disk_block_size);
                                	retstat = cacheLayer->DiskRead(temp_disk_id,temp_buf2,size,offset_read);

					for (k=0; k<size; k++){
						temp_buf[k] = temp_buf[k] ^ 
							(char)gf_mul((unsigned char)coefficient, 
                                                                (unsigned char)temp_buf2[k],field_power);
					}
				}
			}

			coefficient = msr_coding_matrix[code_block_id][msr_block_id];
			for (k=0; k<size; k++){
                        	buf[k] = (char)gf_div((unsigned char)temp_buf[k],
                                		(unsigned char)coefficient, field_power);
			}
		}
		else if ((disk_failed_no == 1) && (flag_code == 1)){ //1 parity node failure. recover parity disk.
                        for (j=0; j<msr_m; j++){
                                coefficient =msr_coding_matrix[code_block_id][j];
                                if (coefficient != 0){
                                        temp_disk_id = msr_get_disk_id(j, msr_segment_size, msr_n, msr_k);
                                        temp_block_no = msr_get_block_no(j, msr_segment_size);
                                        temp_block_no = temp_block_no + msr_segment_id * msr_segment_size;
                                        offset_read = temp_block_no * (NCFS_DATA->disk_block_size);
                                        retstat = cacheLayer->DiskRead(temp_disk_id,temp_buf2,size,offset_read);

                                        for (k=0; k<(NCFS_DATA->disk_block_size); k++){
                                                buf[k] = buf[k] ^ 
							(char)gf_mul((unsigned char)coefficient, 
                                        			(unsigned char)temp_buf2[k],field_power);
                                        }                
                                }
                        }			
		}

                free(temp_buf);
		free(temp_buf2);
                return size;
	}

        AbnormalError();             
        return -1;
}


/*************************************************************************
 * Public functions
 *************************************************************************/

/*
 * Constructor - Initialize Coding_storage layer
 */
CodingLayer::CodingLayer(){

	int coding_type;
	int i;

	coding_type = NCFS_DATA->disk_raid_type;

	if (coding_type == 6){   //raid 6
		gf_gen_tables(field_power);
		//generate GF tables for raid6
	}
	else if (coding_type == 2000){   // MSR exact repair
                //generate GF tables for raid6            
                gf_gen_tables(field_power);

		//generate msr coding matrix
		msr_gen_coding_matrix(NCFS_DATA->msr_n, NCFS_DATA->msr_k,
					NCFS_DATA->msr_m, NCFS_DATA->msr_c);
		for (i=(int)((NCFS_DATA->disk_total_num)/2); i<(int)((NCFS_DATA->disk_total_num)); i++){
			NCFS_DATA->free_offset[i] = NCFS_DATA->free_size[i];
			NCFS_DATA->free_size[i] = 0;
		}
	}
}


/*
 * encode: Write data to disk and return written block info
 *
 * @param buf: data buffer to be written to disk
 * @param size: size of the data buffer 
 *
 * @return: assigned data block: disk id and block number.
 */
struct data_block_info CodingLayer::encode(const char* buf, int size) 
{
	struct data_block_info block_written;
	int disk_raid_type;

	//block_written = default_encoding(buf, size);

	disk_raid_type = NCFS_DATA->disk_raid_type;

	if (disk_raid_type == 100){
		block_written = encoding_jbod(buf, size);
	}
	else if (disk_raid_type == 0){
		block_written = encoding_raid0(buf, size);
	}
	else if (disk_raid_type == 1){
		block_written = encoding_raid1(buf, size);
	}
	else if (disk_raid_type == 4){
		block_written = encoding_raid4(buf, size);
	}
	else if (disk_raid_type == 5){
		block_written = encoding_raid5(buf, size);
	}
        else if (disk_raid_type == 6){
                block_written = encoding_raid6(buf, size);
        }
        else if (disk_raid_type == 1000){
                block_written = encoding_mbr(buf, size);
        }
        else if (disk_raid_type == 2000){
                block_written = encoding_msr(buf, size);
        }
	else{
		block_written = encoding_default(buf, size);
	}

	if (NCFS_DATA->no_gui == 0){
		diskusageLayer->Update(block_written.disk_id,block_written.block_no,1);
	}

	return block_written;
}


/*
 * decode: Select decoding function and read data
 *
 * @param disk_id: disk id
 * @param buf: data buffer for reading data
 * @param size: size of the data buffer
 * @param offset: offset of data on disk
 * 
 * @return: size of successful read (in bytes)
 */
int CodingLayer::decode(int disk_id, char* buf, long long size, long long offset){
	switch(NCFS_DATA->disk_raid_type){
		case 0: return decoding_raid0(disk_id,buf,size,offset);break;
		case 1: return decoding_raid1(disk_id,buf,size,offset);break;
		case 4: return decoding_raid4(disk_id,buf,size,offset);break;
		case 5: return decoding_raid5(disk_id,buf,size,offset);break;
                case 6: return decoding_raid6(disk_id,buf,size,offset);break;
		case 100: return decoding_jbod(disk_id,buf,size,offset);break;
		case 1000: return decoding_mbr(disk_id,buf,size,offset);break;
                case 2000: return decoding_msr(disk_id,buf,size,offset);break;
		default: return decoding_default(disk_id,buf,size,offset);break;
	}
	AbnormalError();
	return -1;
}

